from pathlib import Path

import pytest
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import (
    get_sample_internal_ids_from_sample_sheet,
)
from cg.constants import FileExtensions
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.exc import FlowCellError, MissingFilesError
from cg.meta.demultiplex.validation import (
    is_demultiplexing_complete,
    is_flow_cell_ready_for_delivery,
    validate_demultiplexing_complete,
    validate_flow_cell_delivery_status,
    validate_flow_cell_has_fastq_files,
    validate_sample_sheet_exists,
)
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


def test_is_demultiplexing_complete_true(tmp_path: Path):
    # GIVEN a path where DEMUX_COMPLETE file is present
    (tmp_path / DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()

    # WHEN checking if demultiplexing is complete
    result = is_demultiplexing_complete(tmp_path)

    # THEN the result should be True
    assert result == True


def test_is_demultiplexing_complete_false(tmp_path: Path):
    # GIVEN a path without DEMUX_COMPLETE file

    # WHEN checking if demultiplexing is complete
    result = is_demultiplexing_complete(tmp_path)

    # THEN the result should be False
    assert result == False


def test_is_flow_cell_ready_for_delivery_true(tmp_path: Path):
    # GIVEN a path where DELIVERY file is present
    (tmp_path / DemultiplexingDirsAndFiles.DELIVERY).touch()

    # WHEN checking if the flow cell is ready for delivery
    result = is_flow_cell_ready_for_delivery(tmp_path)

    # THEN the result should be True
    assert result == True


def test_is_flow_cell_ready_for_delivery_false(tmp_path: Path):
    # GIVEN a path without DELIVERY file

    # WHEN checking if the flow cell is ready for delivery
    result = is_flow_cell_ready_for_delivery(tmp_path)

    # THEN the result should be False
    assert result == False


def test_validate_sample_sheet_exists_raises_error(bcl2fastq_flow_cell_dir: Path):
    # GIVEN a flow cell without a sample sheet in housekeeper
    flow_cell = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)
    flow_cell._sample_sheet_path_hk = None
    # WHEN validating the existence of the sample sheet
    # THEN it should raise a FlowCellError
    with pytest.raises(FlowCellError):
        validate_sample_sheet_exists(flow_cell=flow_cell)


def test_validate_sample_sheet_exists(bcl2fastq_flow_cell_dir: Path):
    # GIVEN a path with a sample sheet
    # GIVEN a flow cell without a sample sheet in housekeeper
    flow_cell = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)
    sample_sheet_path = Path(
        bcl2fastq_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )
    sample_sheet_path.touch()
    flow_cell._sample_sheet_path_hk = sample_sheet_path

    # WHEN validating the existence of the sample sheet
    # THEN it should not raise an error
    assert validate_sample_sheet_exists(flow_cell=flow_cell) is None


def test_validate_demultiplexing_complete_raises_error(tmp_path: Path):
    # GIVEN a path without DEMUX_COMPLETE file

    # WHEN validating if demultiplexing is complete
    # THEN it should raise a FlowCellError
    with pytest.raises(FlowCellError):
        validate_demultiplexing_complete(tmp_path)


def test_validate_demultiplexing_complete_no_error(tmp_path: Path):
    # GIVEN a path where DEMUX_COMPLETE file is present
    (tmp_path / DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()

    # WHEN validating if demultiplexing is complete
    # THEN it should not raise an error
    assert validate_demultiplexing_complete(tmp_path) is None


def test_validate_flow_cell_delivery_status_no_error(tmp_path: Path):
    # GIVEN a path without DELIVERY file

    # WHEN validating the flow cell delivery status
    # THEN it should not raise an error
    assert (
        validate_flow_cell_delivery_status(flow_cell_output_directory=tmp_path, force=False) is None
    )


def test_validate_flow_cell_delivery_status_raises_error(tmp_path: Path):
    # GIVEN a path where DELIVERY file is present
    (tmp_path / DemultiplexingDirsAndFiles.DELIVERY).touch()

    # WHEN validating the flow cell delivery status
    # THEN it should raise a FlowCellError
    with pytest.raises(FlowCellError):
        validate_flow_cell_delivery_status(flow_cell_output_directory=tmp_path, force=False)


def test_validate_flow_cell_delivery_status_forced(tmp_path: Path):
    # GIVEN a path where DELIVERY file is present
    (tmp_path / DemultiplexingDirsAndFiles.DELIVERY).touch()

    # WHEN validating the flow cell delivery status
    # THEN it should not raise a FlowCellError
    assert (
        validate_flow_cell_delivery_status(flow_cell_output_directory=tmp_path, force=True) is None
    )


def test_validate_samples_have_fastq_files_passes(
    novaseqx_flow_cell_with_sample_sheet_no_fastq,
):
    """Test the check of a flow cells with one sample fastq file does not raise an error."""
    # GIVEN a demultiplexed flow cell with no fastq files

    # GIVEN a that the flow cell has a sample sheet in Housekeeper
    assert novaseqx_flow_cell_with_sample_sheet_no_fastq.get_sample_sheet_path_hk()

    # GIVEN that a valid sample fastq file is added to the directory
    sample_id: str = get_sample_internal_ids_from_sample_sheet(
        sample_sheet_path=novaseqx_flow_cell_with_sample_sheet_no_fastq.get_sample_sheet_path_hk(),
        flow_cell_sample_type=novaseqx_flow_cell_with_sample_sheet_no_fastq.sample_type,
    )[0]
    fastq_file_path = Path(
        novaseqx_flow_cell_with_sample_sheet_no_fastq.path,
        f"{sample_id}_S11_L1_R1_{FileExtensions.FASTQ}{FileExtensions.GZIP}",
    )
    fastq_file_path.touch(exist_ok=True)

    # WHEN checking if the flow cell has fastq files for the samples
    validate_flow_cell_has_fastq_files(flow_cell=novaseqx_flow_cell_with_sample_sheet_no_fastq)

    # THEN no error is raised


def test_validate_samples_have_fastq_files_fails(
    novaseqx_flow_cell_with_sample_sheet_no_fastq,
):
    """Test the check of a flow cells with one sample fastq file does not raise an error."""
    # GIVEN a demultiplexed flow cell with no fastq files

    # GIVEN a that the flow cell has a sample sheet in Housekeeper
    assert novaseqx_flow_cell_with_sample_sheet_no_fastq.get_sample_sheet_path_hk()

    # WHEN checking if the flow cell has fastq files for the samples
    with pytest.raises(MissingFilesError):
        # THEN an error is raised
        validate_flow_cell_has_fastq_files(flow_cell=novaseqx_flow_cell_with_sample_sheet_no_fastq)



"""Tests for SPRING decompression methods"""
import logging

import pytest
from cg.apps.crunchy import CrunchyAPI
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile, WriteFile


def test_is_spring_decompression_done_all_files_exist(
    crunchy_config: dict, compression_object, spring_metadata_file, caplog
):
    """Test if SPRING decompression is done when FASTQ files are unarchived.

    The function should return true since both FASTQ files and SPRING files exists and there is an
    updated field in the SPRING metadata config.
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the FASTQ paths exists
    compression_object.fastq_first.touch()
    compression_object.fastq_second.touch()
    # GIVEN a existing SPRING file
    compression_object.spring_path.touch()

    # GIVEN a existing flag file
    # GIVEN that the files have an updated tag
    old_date = "2019-01-01"
    content = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # GIVEN that the files where updated more than three weeks ago
    for file_info in content:
        file_info["updated"] = old_date
    WriteFile.write_file_from_content(
        content=content, file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_done(compression_object)

    # THEN result should be True since all files exists and has been updated
    assert result is True
    # THEN assert that it was communicated that decompression is done
    assert f"SPRING decompression is done for run {compression_object.run_name}" in caplog.text


def test_is_spring_decompression_done_missing_fastq_files(
    crunchy_config: dict, compression_object, spring_metadata_file, caplog
):
    """Test if SPRING decompression is done when FASTQ files are missing

    The function should return False since FASTQ files are missing
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the FASTQ paths does not exist
    compression_object.fastq_first.unlink()
    compression_object.fastq_second.unlink()
    assert not compression_object.fastq_first.exists()
    assert not compression_object.fastq_second.exists()

    # GIVEN a existing SPRING file
    compression_object.spring_path.touch()

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_done(compression_object)

    # THEN result should be False since FASTQ files are missing
    assert result is False
    # THEN assert that it was communicated that decompression is done
    assert "does not exist" in caplog.text


def test_is_spring_decompression_done_all_files_exist_not_updated(
    crunchy_config: dict, compression_object, spring_metadata_file, caplog
):
    """Test if SPRING decompression is done when FASTQ files are not unarchived

    The function should return False since the files has not been updated in the metadata file
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the FASTQ paths exists
    compression_object.fastq_first.touch()
    compression_object.fastq_second.touch()
    # GIVEN a existing SPRING file
    compression_object.spring_path.touch()

    # GIVEN a existing flag file
    # GIVEN that the files are missing the updated tag
    content = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # GIVEN that the files where updated more than three weeks ago
    for file_info in content:
        assert "updated" not in file_info

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_done(compression_object)

    # THEN result should be False since all files exists but miss the updated tag
    assert result is False
    # THEN assert that it was communicated that files have not been updated
    assert "Files have not been unarchived" in caplog.text


def test_is_spring_decompression_done_missing_metadata_file(
    crunchy_config: dict, compression_object, caplog
):
    """Test if SPRING decompression is done when SPRING metadata file is missing

    The function should return False since no metadata file is found
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the SPRING metadata file is missing
    assert not compression_object.spring_metadata_path.exists()

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_done(compression_object)

    # THEN result should be False since the metadata file is missing
    assert result is False
    # THEN assert that it was communicated that files have not been updated
    assert "No SPRING metadata file found" in caplog.text


def test_is_spring_decompression_done_empty_metadata_file(
    crunchy_config: dict, compression_object, caplog
):
    """Test if SPRING decompression is done when SPRING metadata file has no content

    The function should return False since the metadata file has no content
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the SPRING metadata file has no content
    compression_object.spring_metadata_path.touch()

    # WHEN checking if SPRING decompression is done
    with pytest.raises(SyntaxError):
        # THEN assert that an exception should be raised since the file is malformed
        crunchy_api.is_spring_decompression_done(compression_object)
        # THEN assert that it was communicated that the content is malformed
        assert "Malformed metadata content" in caplog.text


def test_is_spring_decompression_possible(
    crunchy_config: dict, compression_object, spring_metadata_file
):
    """Test if SPRING decompression is possible when decompression is already done

    The function should return False since decompression is already done
    """
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN that the FASTQ paths exists
    compression_object.fastq_first.touch()
    compression_object.fastq_second.touch()
    # GIVEN a existing SPRING file
    compression_object.spring_path.touch()

    # GIVEN a existing flag file
    # GIVEN that the files have an updated tag
    old_date = "2019-01-01"
    content = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # GIVEN that the files where updated more than three weeks ago
    for file_info in content:
        file_info["updated"] = old_date
    WriteFile.write_file_from_content(
        content=content, file_format=FileFormat.JSON, file_path=spring_metadata_file
    )

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_possible(compression_object)

    # THEN result should be False since decompression is already done
    assert result is False


def test_is_spring_decompression_possible_decompression_pending(
    crunchy_config: dict, compression_object, spring_metadata_file, caplog
):
    """Test if SPRING decompression is possible when decompression is pending

    The function should return False since decompression is pending
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN a existing pending file
    compression_object.pending_path.touch()

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_possible(compression_object)

    # THEN result should be False since decompression is pending
    assert result is False
    # THEN assert that it was communicated that compression is pending
    assert "decompression is pending" in caplog.text


def test_is_spring_decompression_possible(
    crunchy_config: dict, compression_object, spring_metadata_file, caplog
):
    """Test if SPRING decompression is possible

    The function should return True since there is a SPRING file and no FASTQ files
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a crunchy-api
    crunchy_api = CrunchyAPI(crunchy_config)
    # GIVEN a existing SPRING file
    compression_object.spring_path.touch()
    # GIVEN that the FASTQ files does not exist
    compression_object.fastq_first.unlink()
    compression_object.fastq_second.unlink()
    assert not compression_object.fastq_first.exists()
    assert not compression_object.fastq_second.exists()

    # WHEN checking if SPRING decompression is done
    result = crunchy_api.is_spring_decompression_possible(compression_object)

    # THEN result should be True since decompression is possible
    assert result is True
    # THEN assert that it was communicated that compression is pending
    assert "Decompression is possible" in caplog.text



from enum import Enum
from typing import List, Optional, Callable

from sqlalchemy.orm import Query

from cg.store.models import Bed


def get_bed_by_entry_id(beds: Query, bed_entry_id: int, **kwargs) -> Query:
    """Return bed by foreign key."""
    return beds.filter(Bed.id == bed_entry_id)


def get_bed_by_name(beds: Query, bed_name: str, **kwargs) -> Query:
    """Return bed by name."""
    return beds.filter(Bed.name == bed_name)


def get_not_archived_beds(beds: Query, **kwargs) -> Query:
    """Return beds not archived."""
    return beds.filter(Bed.is_archived.is_(False))


def order_beds_by_name(beds: Query, **kwargs) -> Query:
    """Return beds ordered by name."""
    return beds.order_by(Bed.name)


class BedFilter(Enum):
    """Define BED filter functions."""

    FILTER_BY_ENTRY_ID: Callable = get_bed_by_entry_id
    FILTER_BY_NAME: Callable = get_bed_by_name
    FILTER_NOT_ARCHIVED: Callable = get_not_archived_beds
    ORDER_BY_NAME: Callable = order_beds_by_name


def apply_bed_filter(
    beds: Query,
    filter_functions: List[Callable],
    bed_entry_id: Optional[int] = None,
    bed_name: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        beds: Query = function(
            beds=beds,
            bed_entry_id=bed_entry_id,
            bed_name=bed_name,
        )
    return beds



"""Module to parse the BCL convert metrics data into the sequencing statistics model."""

from pathlib import Path
from typing import List
from cg.store.models import SampleLaneSequencingMetrics
from cg.apps.sequencing_metrics_parser.parsers.bcl_convert import BclConvertMetricsParser
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from datetime import datetime
from cg.constants.demultiplexing import BclConverter


def create_sample_lane_sequencing_metrics_from_bcl_convert_metrics_for_flow_cell(
    flow_cell_dir: Path,
) -> List[SampleLaneSequencingMetrics]:
    """Parse the BCL convert metrics data into the sequencing statistics model."""
    metrics_parser: BclConvertMetricsParser = BclConvertMetricsParser(
        bcl_convert_metrics_dir_path=flow_cell_dir,
    )
    sample_internal_ids: List[str] = metrics_parser.get_sample_internal_ids()
    sample_lane_sequencing_metrics: List[SampleLaneSequencingMetrics] = []
    for sample_internal_id in sample_internal_ids:
        for lane in metrics_parser.get_lanes_for_sample(sample_internal_id=sample_internal_id):
            sample_lane_sequencing_metrics.append(
                SampleLaneSequencingMetrics(
                    sample_internal_id=sample_internal_id,
                    flow_cell_name=FlowCellDirectoryData(
                        flow_cell_path=flow_cell_dir, bcl_converter=BclConverter.BCLCONVERT
                    ).id,
                    flow_cell_lane_number=lane,
                    sample_total_reads_in_lane=metrics_parser.calculate_total_reads_for_sample_in_lane(
                        sample_internal_id=sample_internal_id, lane=lane
                    ),
                    sample_base_percentage_passing_q30=metrics_parser.get_q30_bases_percent_for_sample_in_lane(
                        sample_internal_id=sample_internal_id, lane=lane
                    ),
                    sample_base_mean_quality_score=metrics_parser.get_mean_quality_score_for_sample_in_lane(
                        sample_internal_id=sample_internal_id, lane=lane
                    ),
                    created_at=datetime.now(),
                )
            )

    return sample_lane_sequencing_metrics



"""Tests for running the delivery command"""

import logging
from pathlib import Path

from cg.cli.deliver.base import deliver_analysis, deliver_ticket
from cg.constants.process import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner


def test_run_deliver_with_help(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN a cli runner and a base context
    # WHEN running the deliver command with help text
    result = cli_runner.invoke(deliver_analysis, ["--help"], obj=base_context)

    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert that the help text is displayed
    assert "Deliver analysis files to customer inbox" in result.output


def test_run_deliver_without_specifying_case_or_ticket(
    cli_runner: CliRunner, base_context: CGConfig, caplog
):
    caplog.set_level(logging.INFO)
    # GIVEN a cli runner and a base context
    # WHEN running the deliver command with help text
    result = cli_runner.invoke(deliver_analysis, ["--delivery-type", "mip-dna"], obj=base_context)

    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert that the help text is displayed
    assert "Please provide a case-id or ticket-id" in caplog.text


def test_run_deliver_non_existing_case(
    cli_runner: CliRunner, base_context: CGConfig, case_id: str, caplog
):
    """Test to run the deliver command when the provided case does not exist"""
    caplog.set_level(logging.WARNING)
    # GIVEN a cli runner and a base context
    # GIVEN a case_id that does not exist in the database
    store: Store = base_context.status_db
    assert store.get_case_by_internal_id(internal_id=case_id) is None

    # WHEN running the deliver command with the non existing case
    result = cli_runner.invoke(
        deliver_analysis, ["--case-id", case_id, "--delivery-type", "mip-dna"], obj=base_context
    )

    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert that the correct warning is displayed
    assert f"Could not find case {case_id}" in caplog.text


def test_delivery_with_dry_run(
    cli_runner: CliRunner,
    populated_mip_context: CGConfig,
    case_id: str,
    delivery_inbox: Path,
    caplog,
):
    """Test to run the delivery command with dry run enabled"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a cli runner
    # GIVEN a context with a case that have files in housekeeper to deliver
    # GIVEN that the delivery path does not exist
    assert delivery_inbox.exists() is False

    # WHEN running the deliver analysis command in dry run mode
    cli_runner.invoke(
        deliver_analysis,
        ["--case-id", case_id, "--delivery-type", "mip-dna", "--dry-run"],
        obj=populated_mip_context,
    )

    # THEN assert that the path to the delivery folder was not created
    assert delivery_inbox.exists() is False

    # THEN assert that it is communicated that at least one file would be linked
    assert "Would hard link file" in caplog.text


def test_delivery_path_created(
    cli_runner: CliRunner, populated_mip_context: CGConfig, case_id: str, delivery_inbox: Path
):
    """Test that the delivery path is created when running the deliver analysis command"""
    # GIVEN a context with a case that have files in housekeeper to deliver
    # GIVEN a cli runner
    # GIVEN that the delivery file does not exist
    assert delivery_inbox.exists() is False

    # WHEN running the deliver analysis command
    cli_runner.invoke(
        deliver_analysis,
        ["--case-id", case_id, "--delivery-type", "mip-dna"],
        obj=populated_mip_context,
    )

    # THEN assert that the path to the delivery folder was created
    assert delivery_inbox.exists() is True


def test_delivery_ticket_id(
    cli_runner: CliRunner,
    populated_mip_context: CGConfig,
    delivery_inbox: Path,
    ticket_id: str,
):
    """Test that to run the deliver command with ticket nr"""
    # GIVEN a context with a case that have files in housekeeper to deliver
    # GIVEN a cli runner
    # GIVEN that the delivery file does not exist
    assert delivery_inbox.exists() is False

    # WHEN running the deliver analysis command
    cli_runner.invoke(
        deliver_analysis,
        ["--ticket", ticket_id, "--delivery-type", "mip-dna"],
        obj=populated_mip_context,
    )

    # THEN assert that the path to the delivery folder was created
    assert delivery_inbox.exists() is True


def test_run_deliver_multiple_delivery_flags(
    cli_runner: CliRunner,
    populated_mip_context: CGConfig,
    case_id: str,
    deliver_vcf_path: Path,
    deliver_fastq_path: Path,
    caplog,
):
    """Test to run the deliver command when the provided case does not exist"""
    caplog.set_level(logging.WARNING)
    # GIVEN a context with a case that has files and a sample with a file, in housekeeper to deliver
    # GIVEN a cli runner
    # WHEN running the deliver command with multiple delivery flags

    assert deliver_vcf_path.exists() is False
    assert deliver_fastq_path.exists() is False

    result = cli_runner.invoke(
        deliver_analysis,
        ["--case-id", case_id, "--delivery-type", "fastq", "--delivery-type", "mip-dna"],
        obj=populated_mip_context,
    )

    # THEN assert the command exits without problems
    assert result.exit_code == 0
    # THEN assert that the case file was delivered to the inbox
    assert deliver_vcf_path.exists() is True
    # THEN assert that the sample file was delivered to the inbox
    assert deliver_fastq_path.exists() is True


def test_case_file_is_delivered(
    populated_mip_context: CGConfig, case_id: str, deliver_vcf_path: Path, cli_runner: CliRunner
):
    """Test that the a case file is delivered when running the delivery command"""
    # GIVEN a context with a case that have files in housekeeper to deliver
    # GIVEN a cli runner
    # GIVEN that a case vcf does not exist
    assert deliver_vcf_path.exists() is False

    # WHEN running the deliver analysis command
    cli_runner.invoke(
        deliver_analysis,
        ["--case-id", case_id, "--delivery-type", "mip-dna"],
        obj=populated_mip_context,
    )

    # THEN assert that the case file was delivered to the inbox
    assert deliver_vcf_path.exists() is True


def test_delivering_analysis_with_missing_bundle_errors(
    cli_runner: CliRunner,
    context_with_missing_bundle: CGConfig,
    ticket_id: str,
):
    """Test that the deliver command fails when a bundle is missing."""
    # GIVEN a context with a case that does not have files in housekeeper to deliver.
    # GIVEN a cli runner
    # WHEN running the deliver analysis command
    result = cli_runner.invoke(
        deliver_analysis,
        ["--ticket", ticket_id, "--delivery-type", "mip-dna"],
        obj=context_with_missing_bundle,
    )

    # THEN assert that the command failed
    assert result.exit_code is not EXIT_SUCCESS


def test_delivering_analysis_with_missing_bundle_ignoring_errors(
    cli_runner: CliRunner,
    context_with_missing_bundle: CGConfig,
    delivery_inbox: Path,
    ticket_id: str,
):
    """Test that it is possible to deliver analysis with a missing bundle using the --ignore-missing-bundles flag."""
    # GIVEN a context without files in housekeeper to deliver.
    # GIVEN a cli runner
    # GIVEN that the delivery file does not exist
    assert delivery_inbox.exists() is False

    # WHEN running the deliver analysis command
    cli_runner.invoke(
        deliver_analysis,
        ["--ticket", ticket_id, "--ignore-missing-bundles", "--delivery-type", "mip-dna"],
        obj=context_with_missing_bundle,
    )

    # THEN assert that the path to the delivery folder was created
    assert delivery_inbox.exists() is True


def test_deliver_ticket_with_missing_bundle(
    cli_runner: CliRunner, context_with_missing_bundle: CGConfig, caplog, ticket_id
):
    caplog.set_level(logging.INFO)

    # GIVEN a cli runner
    # WHEN running cg deliver ticket
    result = cli_runner.invoke(
        deliver_ticket,
        [
            "--ticket",
            ticket_id,
            "--dry-run",
            "--ignore-missing-bundles",
            "--delivery-type",
            "mip-dna",
        ],
        obj=context_with_missing_bundle,
    )

    # THEN assert that the command succeeded and files are delivered
    assert result.exit_code is EXIT_SUCCESS
    assert "Delivering files to customer inbox on the HPC" in caplog.text



"""Module to handle time functions"""

import time


def get_start_time() -> float:
    """Returns the current time"""
    return time.time()


def get_elapsed_time(start_time: float) -> float:
    """Determines the time elapsed since the start time"""
    return time.time() - start_time



""" Create a sample sheet for NovaSeq flow cells."""
import logging
from typing import Dict, List, Optional, Set, Type, Union

from cg.apps.demultiplex.sample_sheet.dummy_sample import get_dummy_sample
from cg.apps.demultiplex.sample_sheet.index import (
    Index,
    get_index_pair,
    get_indexes_by_lane,
    get_valid_indexes,
    index_exists,
    is_dual_index,
    is_reverse_complement_needed,
    update_barcode_mismatch_values_for_sample,
    update_indexes_for_samples,
)
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSample,
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import (
    get_samples_by_lane,
    get_validated_sample_sheet,
)
from cg.constants.demultiplexing import (
    BclConverter,
    SampleSheetBcl2FastqSections,
    SampleSheetBCLConvertSections,
)
from cg.exc import SampleSheetError
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.models.demultiplex.run_parameters import RunParameters

LOG = logging.getLogger(__name__)


class SampleSheetCreator:
    """Base class for sample sheet creation."""

    def __init__(
        self,
        flow_cell: FlowCellDirectoryData,
        lims_samples: List[Union[FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq]],
        force: bool = False,
    ):
        self.flow_cell: FlowCellDirectoryData = flow_cell
        self.flow_cell_id: str = flow_cell.id
        self.lims_samples: List[
            Union[FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq]
        ] = lims_samples
        self.run_parameters: RunParameters = flow_cell.run_parameters
        self.sample_type: Type[
            Union[FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq]
        ] = flow_cell.sample_type
        self.force: bool = force

    @property
    def bcl_converter(self) -> str:
        """Return the bcl converter used for demultiplexing."""
        return self.flow_cell.bcl_converter

    @property
    def valid_indexes(self) -> List[Index]:
        return get_valid_indexes(dual_indexes_only=True)

    def add_dummy_samples(self) -> None:
        """Add all dummy samples with non-existing indexes to samples if applicable."""
        raise NotImplementedError("Impossible to add dummy samples in parent class")

    def update_barcode_mismatch_values_for_samples(self, *args) -> None:
        """Updates barcode mismatch values for samples if applicable."""
        raise NotImplementedError(
            "Impossible to update sample barcode mismatches from parent class"
        )

    def add_override_cycles_to_samples(self) -> None:
        """Add override cycles attribute to samples if sample sheet is v2."""
        raise NotImplementedError("Impossible to add override cycles to samples from parent class")

    def remove_unwanted_samples(self) -> None:
        """Filter out samples with single indexes."""
        LOG.info("Removing all samples without dual indexes")
        samples_to_keep = []
        sample: Union[FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq]
        for sample in self.lims_samples:
            if not is_dual_index(sample.index):
                LOG.warning(f"Removing sample {sample} since it does not have dual index")
                continue
            samples_to_keep.append(sample)
        self.lims_samples = samples_to_keep

    @staticmethod
    def convert_sample_to_header_dict(
        sample: Union[FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq],
        data_column_names: List[str],
    ) -> List[str]:
        """Convert a lims sample object to a list that corresponds to the sample sheet headers."""
        LOG.debug(f"Use sample sheet header {data_column_names}")
        sample_dict = sample.model_dump(by_alias=True)
        return [str(sample_dict[column]) for column in data_column_names]

    def get_additional_sections_sample_sheet(self) -> Optional[List]:
        """Return all sections of the sample sheet that are not the data section."""
        raise NotImplementedError("Impossible to get sample sheet sections from parent class")

    def get_data_section_header_and_columns(self) -> Optional[List[List[str]]]:
        """Return the header and column names of the data section of the sample sheet."""
        raise NotImplementedError("Impossible to get sample sheet sections from parent class")

    def create_sample_sheet_content(self) -> List[List[str]]:
        """Create sample sheet content with samples."""
        LOG.info("Creating sample sheet content")
        sample_sheet_content: List[List[str]] = (
            self.get_additional_sections_sample_sheet() + self.get_data_section_header_and_columns()
        )
        for sample in self.lims_samples:
            sample_sheet_content.append(
                self.convert_sample_to_header_dict(
                    sample=sample,
                    data_column_names=self.get_data_section_header_and_columns()[1],
                )
            )
        return sample_sheet_content

    def process_samples_for_sample_sheet(self) -> None:
        """Add dummy samples, remove unwanted samples and adapt remaining samples."""
        if self.run_parameters.requires_dummy_samples:
            self.add_dummy_samples()
            LOG.info("Created dummy samples for the indexes that are missing")
        else:
            LOG.info("Skipped adding dummy samples since they are not needed")
        self.remove_unwanted_samples()
        samples_in_lane: List[Union[FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq]]
        is_reverse_complement: bool = is_reverse_complement_needed(
            run_parameters=self.run_parameters
        )
        self.add_override_cycles_to_samples()
        for lane, samples_in_lane in get_samples_by_lane(self.lims_samples).items():
            LOG.info(f"Adapting index and barcode mismatch values for samples in lane {lane}")
            update_indexes_for_samples(
                samples=samples_in_lane,
                index_cycles=self.run_parameters.index_length,
                is_reverse_complement=is_reverse_complement,
            )
            self.update_barcode_mismatch_values_for_samples(samples_in_lane)

    def construct_sample_sheet(self) -> List[List[str]]:
        """Construct and validate the sample sheet."""
        self.process_samples_for_sample_sheet()
        sample_sheet_content: List[List[str]] = self.create_sample_sheet_content()
        if self.force:
            LOG.info("Skipping validation of sample sheet due to force flag")
            return sample_sheet_content
        LOG.info("Validating sample sheet")
        get_validated_sample_sheet(
            sample_sheet_content=sample_sheet_content,
            sample_type=self.sample_type,
        )
        LOG.info("Sample sheet passed validation")
        return sample_sheet_content


class SampleSheetCreatorBcl2Fastq(SampleSheetCreator):
    """Create a raw sample sheet for flow cells."""

    def add_dummy_samples(self) -> None:
        """Add all dummy samples with non-existing indexes to samples.

        Dummy samples are added if there are indexes that are not used by the actual samples.
        """
        LOG.info("Adding dummy samples for unused indexes")
        indexes_by_lane: Dict[int, Set[str]] = get_indexes_by_lane(samples=self.lims_samples)
        for lane, lane_indexes in indexes_by_lane.items():
            LOG.debug(f"Add dummy samples for lane {lane}")
            for index in self.valid_indexes:
                if index_exists(index=index.sequence, indexes=lane_indexes):
                    LOG.debug(f"Index {index.sequence} already in use")
                    continue
                dummy_flow_cell_sample: Union[
                    FlowCellSampleBCLConvert, FlowCellSampleBcl2Fastq
                ] = get_dummy_sample(
                    flow_cell_id=self.flow_cell_id,
                    dummy_index=index.sequence,
                    lane=lane,
                    name=index.name,
                    sample_type=self.sample_type,
                )
                LOG.debug(f"Adding dummy sample {dummy_flow_cell_sample} to lane {lane}")
                self.lims_samples.append(dummy_flow_cell_sample)

    def update_barcode_mismatch_values_for_samples(self, *args) -> None:
        """Return None for flow cells to be demultiplexed with Bcl2fastq."""
        LOG.debug("No barcode mismatch updating for Bcl2fastq flow cell")

    def add_override_cycles_to_samples(self) -> None:
        """Return None for flow cells to be demultiplexed with Bcl2fastq."""
        LOG.debug("No adding of override cycles for Bcl2fastq flow cell")

    def get_additional_sections_sample_sheet(self) -> List[List[str]]:
        """Return all sections of the sample sheet that are not the data section."""
        return [
            [SampleSheetBcl2FastqSections.Settings.HEADER.value],
            SampleSheetBcl2FastqSections.Settings.BARCODE_MISMATCH_INDEX1.value,
            SampleSheetBcl2FastqSections.Settings.BARCODE_MISMATCH_INDEX2.value,
        ]

    def get_data_section_header_and_columns(self) -> List[List[str]]:
        """Return the header and column names of the data section of the sample sheet."""
        return [
            [SampleSheetBcl2FastqSections.Data.HEADER.value],
            SampleSheetBcl2FastqSections.Data.COLUMN_NAMES.value,
        ]


class SampleSheetCreatorBCLConvert(SampleSheetCreator):
    """Create a raw sample sheet for BCLConvert flow cells."""

    def __init__(
        self,
        flow_cell: FlowCellDirectoryData,
        lims_samples: List[FlowCellSampleBCLConvert],
        force: bool = False,
    ):
        super().__init__(flow_cell, lims_samples, force)
        if flow_cell.bcl_converter == BclConverter.BCL2FASTQ:
            raise SampleSheetError(f"Can't use {BclConverter.BCL2FASTQ} with sample sheet v2")

    def add_dummy_samples(self) -> None:
        """Return None for flow cells to be demultiplexed with BCLConvert."""
        LOG.debug("No adding of dummy samples for for BCLConvert flow cell")

    def update_barcode_mismatch_values_for_samples(
        self, samples: List[FlowCellSampleBCLConvert]
    ) -> None:
        """Update barcode mismatch values for both indexes of given samples."""
        for sample in samples:
            update_barcode_mismatch_values_for_sample(
                sample_to_update=sample, samples_to_compare_to=samples
            )

    def add_override_cycles_to_samples(self) -> None:
        """Add override cycles attribute to samples."""
        flow_cell_index_len: int = self.run_parameters.index_length
        read1_cycles: str = f"Y{self.run_parameters.get_read_1_cycles()};"
        read2_cycles: str = f"Y{self.run_parameters.get_read_2_cycles()}"
        for sample in self.lims_samples:
            index1_cycles: str = f"I{self.run_parameters.get_index_1_cycles()};"
            index2_cycles: str = f"I{self.run_parameters.get_index_2_cycles()};"
            sample_index_len: int = len(get_index_pair(sample)[0])
            if sample_index_len < flow_cell_index_len:
                index1_cycles = f"I{sample_index_len}N{flow_cell_index_len - sample_index_len};"
                index2_cycles = f"N{flow_cell_index_len - sample_index_len}I{sample_index_len};"
            sample.override_cycles = read1_cycles + index1_cycles + index2_cycles + read2_cycles

    def get_additional_sections_sample_sheet(self) -> List[List[str]]:
        """Return all sections of the sample sheet that are not the data section."""
        header_section: List[List[str]] = [
            [SampleSheetBCLConvertSections.Header.HEADER.value],
            SampleSheetBCLConvertSections.Header.FILE_FORMAT.value,
            [SampleSheetBCLConvertSections.Header.RUN_NAME.value, self.flow_cell_id],
            [
                SampleSheetBCLConvertSections.Header.INSTRUMENT_PLATFORM_TITLE.value,
                SampleSheetBCLConvertSections.Header.INSTRUMENT_PLATFORM_VALUE.value[
                    self.flow_cell.sequencer_type
                ],
            ],
            SampleSheetBCLConvertSections.Header.INDEX_ORIENTATION_FORWARD.value,
        ]
        reads_section: List[List[str]] = [
            [SampleSheetBCLConvertSections.Reads.HEADER.value],
            [
                SampleSheetBCLConvertSections.Reads.READ_CYCLES_1.value,
                self.run_parameters.get_read_1_cycles(),
            ],
            [
                SampleSheetBCLConvertSections.Reads.READ_CYCLES_2.value,
                self.run_parameters.get_read_2_cycles(),
            ],
            [
                SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_1.value,
                self.run_parameters.get_index_1_cycles(),
            ],
            [
                SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2.value,
                self.run_parameters.get_index_2_cycles(),
            ],
        ]
        settings_section: List[List[str]] = [
            [SampleSheetBCLConvertSections.Settings.HEADER.value],
            SampleSheetBCLConvertSections.Settings.SOFTWARE_VERSION.value,
            SampleSheetBCLConvertSections.Settings.FASTQ_COMPRESSION_FORMAT.value,
        ]
        return header_section + reads_section + settings_section

    def get_data_section_header_and_columns(self) -> List[List[str]]:
        """Return the header and column names of the data section of the sample sheet."""
        return [
            [SampleSheetBCLConvertSections.Data.HEADER.value],
            SampleSheetBCLConvertSections.Data.COLUMN_NAMES.value,
        ]



"""Fixtures for cli Compress functions."""

import datetime as dt
from datetime import datetime
from pathlib import Path
from typing import Dict, Any, Generator

import pytest
from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI

from cg.constants import FileExtensions, SequencingFileTag
from cg.constants.pedigree import Pedigree
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from tests.store_helpers import StoreHelpers


class MockCompressAPI(CompressAPI):
    """Mock out necessary functions for running the compress CLI functions."""

    def __init__(self):
        """initialize mock."""
        super().__init__(hk_api=None, crunchy_api=None, demux_root="")
        self.fastq_compression_success: bool = True
        self.spring_decompression_success: bool = True
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Update dry run."""
        self.dry_run = dry_run

    def compress_fastq(self, sample_id: str, dry_run: bool = False) -> None:
        """Return if compression was successful."""
        _ = sample_id, dry_run
        return self.fastq_compression_success

    def decompress_spring(self, sample_id: str, dry_run: bool = False) -> None:
        """Return if decompression was successful."""
        _ = sample_id, dry_run
        return self.spring_decompression_success


@pytest.fixture
def compress_api() -> MockCompressAPI:
    """Return a Compress context."""
    return MockCompressAPI()


@pytest.fixture
def real_crunchy_api(
    crunchy_config: Dict[str, Dict[str, Any]]
) -> Generator[CrunchyAPI, None, None]:
    """Return Crunchy API."""
    _api = CrunchyAPI(crunchy_config)
    _api.set_dry_run(True)
    yield _api


@pytest.fixture
def real_compress_api(
    demultiplex_runs: Path, housekeeper_api: HousekeeperAPI, real_crunchy_api: CrunchyAPI
) -> CompressAPI:
    """Return a Compress context."""
    return CompressAPI(
        crunchy_api=real_crunchy_api, hk_api=housekeeper_api, demux_root=demultiplex_runs.as_posix()
    )


@pytest.fixture
def real_populated_compress_fastq_api(
    real_compress_api: CompressAPI, compress_hk_fastq_bundle: dict, helpers: StoreHelpers
) -> CompressAPI:
    """Return populated Compress API."""
    helpers.ensure_hk_bundle(real_compress_api.hk_api, compress_hk_fastq_bundle)
    return real_compress_api


# Store fixtures


class CaseInfo:
    """Holds information for creating a case."""

    def __init__(self, **kwargs):
        self.case_id: str = kwargs["case_id"]
        self.family_name: str = kwargs["family_name"]
        self.timestamp: datetime = kwargs["timestamp"]
        self.later_timestamp: datetime = kwargs["later_timestamp"]
        self.application_tag: str = kwargs["application_tag"]


@pytest.fixture
def compress_case_info(
    case_id: str,
    family_name: str,
    timestamp: datetime,
    later_timestamp: datetime,
    wgs_application_tag: str,
) -> CaseInfo:
    """Returns an object with information about a case."""
    return CaseInfo(
        case_id=case_id,
        family_name=family_name,
        timestamp=timestamp,
        later_timestamp=later_timestamp,
        application_tag=wgs_application_tag,
    )


@pytest.fixture
def populated_compress_store(
    store: Store, helpers: StoreHelpers, compress_case_info, analysis_family
):
    """Return a store populated with a completed analysis."""
    helpers.ensure_case_from_dict(
        store,
        case_info=analysis_family,
        app_tag=compress_case_info.application_tag,
        ordered_at=compress_case_info.timestamp,
        completed_at=compress_case_info.later_timestamp,
    )
    return store


@pytest.fixture
def populated_compress_multiple_store(
    store: Store,
    helpers: StoreHelpers,
    compress_case_info: CaseInfo,
    analysis_family: dict,
) -> Store:
    """Return a store populated with multiple completed analyses."""
    case_id: str = compress_case_info.case_id
    family_name: str = compress_case_info.family_name
    for number in range(10):
        analysis_family["internal_id"]: str = "_".join([str(number), case_id])
        analysis_family["name"]: str = "_".join([str(number), family_name])
        for ind, sample in enumerate(analysis_family["samples"]):
            analysis_family["samples"][ind]["internal_id"]: str = "_".join(
                [str(number), sample["internal_id"]]
            )
            if Pedigree.FATHER in analysis_family["samples"][ind]:
                analysis_family["samples"][ind][Pedigree.FATHER]: str = "_".join(
                    [str(number), analysis_family["samples"][ind][Pedigree.FATHER]]
                )
            if Pedigree.MOTHER in analysis_family["samples"][ind]:
                analysis_family["samples"][ind][Pedigree.MOTHER]: str = "_".join(
                    [str(number), analysis_family["samples"][ind][Pedigree.MOTHER]]
                )

        helpers.ensure_case_from_dict(
            store,
            case_info=analysis_family,
            app_tag=compress_case_info.application_tag,
            ordered_at=compress_case_info.timestamp,
            completed_at=compress_case_info.later_timestamp,
            created_at=dt.datetime.now() - dt.timedelta(days=1000),
        )
    return store


# Context fixtures
@pytest.fixture
def base_compress_context(
    compress_api: CompressAPI, store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a Compress context."""
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = store
    return cg_config_object


@pytest.fixture
def populated_multiple_compress_context(
    compress_api: CompressAPI, populated_compress_multiple_store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a Compress context populated with a completed analysis."""
    cg_config_object.meta_apis["compress_api"]: CompressAPI = compress_api
    cg_config_object.status_db_ = populated_compress_multiple_store
    return cg_config_object


@pytest.fixture
def populated_compress_context(
    compress_api: CompressAPI, populated_compress_store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a Compress context populated with a completed analysis."""
    cg_config_object.meta_apis["compress_api"]: CompressAPI = compress_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object


@pytest.fixture
def real_populated_compress_context(
    real_populated_compress_fastq_api: CompressAPI,
    populated_compress_store: Store,
    cg_config_object: CGConfig,
) -> CGConfig:
    """Return a Compress context populated with a completed analysis."""
    cg_config_object.meta_apis["compress_api"]: CompressAPI = real_populated_compress_fastq_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object


# Bundle fixtures


@pytest.fixture
def sample() -> str:
    """Return the sample id for first sample."""
    return "sample_1"


@pytest.fixture
def new_dir(project_dir: Path) -> Path:
    """Return the path to a subdirectory."""
    new_dir = Path(project_dir, "new_dir")
    new_dir.mkdir()
    return new_dir


@pytest.fixture
def spring_bundle(project_dir: Path, timestamp: datetime, sample: str) -> dict:
    """Return a bundle with spring files."""
    spring_file: Path = Path(project_dir, f"file{FileExtensions.SPRING}")
    spring_file.touch()
    spring_meta_file: Path = Path(project_dir, f"file{FileExtensions.JSON}")
    spring_meta_file.touch()
    return {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(spring_file),
                "archive": False,
                "tags": [sample, SequencingFileTag.SPRING],
            },
            {
                "path": str(spring_meta_file),
                "archive": False,
                "tags": [sample, SequencingFileTag.SPRING_METADATA],
            },
        ],
    }


@pytest.fixture
def spring_bundle_symlink_problem(
    project_dir: Path, new_dir: Path, timestamp: datetime, sample: str
):
    """Return Housekeeper bundle with SPRING files having symlinks."""
    spring_file: Path = Path(project_dir, "file.spring")
    spring_file.touch()
    wrong_spring_file: Path = Path(new_dir, "file.spring")
    spring_meta_file: Path = Path(project_dir, "file.json")
    spring_meta_file.touch()
    wrong_spring_meta_file: Path = Path(new_dir, "file.json")
    return {
        "name": sample,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(wrong_spring_file),
                "archive": False,
                "tags": [sample, SequencingFileTag.SPRING],
            },
            {
                "path": str(wrong_spring_meta_file),
                "archive": False,
                "tags": [sample, SequencingFileTag.SPRING_METADATA],
            },
        ],
    }


@pytest.fixture
def symlinked_fastqs(project_dir: Path, new_dir: Path) -> dict:
    """Setup an environment that is similar to the case we want to solve."""
    fastq_first: Path = Path(project_dir, "first.fastq.gz")
    fastq_second: Path = Path(project_dir, "second.fastq.gz")

    symlinked_first: Path = Path(new_dir, "first.fastq.gz")
    symlinked_first.symlink_to(fastq_first)
    symlinked_second: Path = Path(new_dir, "second.fastq.gz")
    symlinked_second.symlink_to(fastq_second)
    return {
        "fastq_first": fastq_first,
        "fastq_second": fastq_second,
        "symlinked_first": symlinked_first,
        "symlinked_second": symlinked_second,
    }



"""Fixtures for cli analysis tests"""
from datetime import datetime
from pathlib import Path

import pytest
from cg.constants import Pipeline, DataDelivery
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family
from tests.store_helpers import StoreHelpers


@pytest.fixture
def base_context(cg_context: CGConfig, analysis_store: Store) -> CGConfig:
    """context to use in cli"""
    cg_context.status_db_ = analysis_store
    return cg_context


@pytest.fixture
def workflow_case_id() -> str:
    """Return a special case id"""
    return "dna_case"


@pytest.fixture(scope="function")
def analysis_store(base_store: Store, workflow_case_id: str, helpers: StoreHelpers) -> Store:
    """Store to be used in tests"""
    _store = base_store

    case = helpers.add_case(_store, workflow_case_id, data_analysis=Pipeline.MIP_DNA)

    dna_sample = helpers.add_sample(
        _store, "dna_sample", is_rna=False, sequenced_at=datetime.now(), reads=10000000
    )
    helpers.add_relationship(_store, sample=dna_sample, case=case)

    case = helpers.add_case(_store, "rna_case", data_analysis=Pipeline.MIP_RNA)
    rna_sample = helpers.add_sample(
        _store, "rna_sample", is_rna=True, sequenced_at=datetime.now(), reads=10000000
    )
    helpers.add_relationship(_store, sample=rna_sample, case=case)

    case = helpers.add_case(_store, "dna_rna_mix_case", data_analysis=Pipeline.MIP_DNA)
    helpers.add_relationship(_store, sample=rna_sample, case=case)
    helpers.add_relationship(_store, sample=dna_sample, case=case)

    return _store


@pytest.fixture
def fastq_context(
    base_context,
    cg_context: CGConfig,
    fastq_case,
    helpers: StoreHelpers,
    case_id,
) -> CGConfig:
    """Returns a CGConfig where the meta_apis["analysis_api"] is a FastqAnalysisAPI and a store
    containing a fastq case"""
    _store = cg_context.status_db
    # Add fastq case to db
    fastq_case["samples"][0]["sequenced_at"] = datetime.now()
    helpers.ensure_case_from_dict(store=_store, case_info=fastq_case)
    return cg_context


@pytest.fixture
def fastq_case(case_id, family_name, sample_id, cust_sample_id, ticket_id: str) -> dict:
    """Returns a dict describing a fastq case"""
    return {
        "name": family_name,
        "panels": None,
        "internal_id": case_id,
        "data_analysis": Pipeline.FASTQ,
        "data_delivery": DataDelivery.FASTQ,
        "completed_at": None,
        "action": None,
        "tickets": ticket_id,
        "samples": [
            {
                "internal_id": sample_id,
                "sex": "male",
                "name": cust_sample_id,
                "original_ticket": ticket_id,
                "reads": 1000000,
                "capture_kit": "anything",
            },
        ],
    }


@pytest.fixture(scope="function")
def dna_case(analysis_store, helpers) -> Family:
    """Case with DNA application"""
    cust = helpers.ensure_customer(analysis_store)
    return analysis_store.get_case_by_name_and_customer(customer=cust, case_name="dna_case")


@pytest.fixture(scope="function")
def rna_case(analysis_store, helpers) -> Family:
    """Case with RNA application"""
    cust = helpers.ensure_customer(analysis_store)
    return analysis_store.get_case_by_name_and_customer(customer=cust, case_name="rna_case")


@pytest.fixture(scope="function")
def dna_rna_mix_case(analysis_store, helpers) -> Family:
    """Case with MIP analysis type DNA and RNA application"""
    cust = helpers.ensure_customer(analysis_store)
    return analysis_store.get_case_by_name_and_customer(customer=cust, case_name="dna_rna_mix_case")


@pytest.fixture
def create_multiqc_html_file(tmpdir_factory) -> Path:
    output_dir = tmpdir_factory.mktemp("output")
    file_path = Path(output_dir, "multiqc_report.html")
    file_path.touch(exist_ok=True)
    return file_path


@pytest.fixture
def create_multiqc_json_file(tmpdir_factory) -> Path:
    output_dir = tmpdir_factory.mktemp("output")
    file_path = Path(output_dir, "multiqc_report.json")
    file_path.touch(exist_ok=True)
    return file_path


class MockTB:
    """Trailblazer mock fixture"""

    def __init__(self):
        self._link_was_called = False
        self._mark_analyses_deleted_called = False
        self._add_pending_was_called = False
        self._add_pending_analysis_was_called = False
        self._family = None
        self._temp = None
        self._case_id = None
        self._email = None
        self._status = None

    def analyses(
        self,
        family=None,
        status=None,
        temp=None,
        case_id=None,
        query: str = None,
        deleted: bool = None,
        before=None,
        is_visible: bool = None,
    ):
        """Mock TB analyses models"""

        self._family = family
        self._status = status
        self._temp = temp

        if case_id == "yellowhog":
            return []

        class Row:
            """Mock a record representing an analysis"""

            def __init__(self):
                """We need to initialize _first_was_called
                so that we can set it in `first()` and retrieve
                it in `first_was_called()`. This way we can easily
                run the invoking code and make sure the function was
                called.
                """

                self._first_was_called = False

            def first(self):
                """Mock that the first row doesn't exist"""
                self._first_was_called = True

            def first_was_called(self):
                """Check if first was called"""
                return self._first_was_called

        return Row()

    def mark_analyses_deleted(self, case_id: str):
        """Mock this function"""
        self._case_id = case_id
        self._mark_analyses_deleted_called = True

    def add_pending(self, case_id: str, email: str):
        """Mock this function"""
        self._case_id = case_id
        self._email = email
        self._add_pending_was_called = True

    def add_pending_analysis(self, case_id: str, email: str):
        """Mock adding a pending analyses"""
        self._case_id = case_id
        self._email = email
        self._add_pending_analysis_was_called = True

    def mark_analyses_deleted_called(self):
        """check if mark_analyses_deleted was called"""
        return self._mark_analyses_deleted_called

    def add_pending_was_called(self):
        """check if add_pending was called"""
        return self._add_pending_was_called

    def is_latest_analysis_ongoing(self, case_id: str):
        """Override TrailblazerAPI is_ongoing method to avoid default behaviour"""
        return False

    def is_latest_analysis_failed(self, case_id: str):
        """Override TrailblazerAPI is_failed method to avoid default behaviour"""
        return False

    def is_latest_analysis_completed(self, case_id: str):
        """Override TrailblazerAPI is_completed method to avoid default behaviour"""
        return False

    def get_latest_analysis_status(self, case_id: str):
        """Override TrailblazerAPI get_analysis_status method to avoid default behaviour"""
        return None

    def has_latest_analysis_started(self, case_id: str):
        """Override TrailblazerAPI has_analysis_started method to avoid default behaviour"""
        return False

    def set_analysis_uploaded(self, case_id: str, uploaded_at: datetime):
        return None


@pytest.fixture(scope="function")
def tb_api():
    """Trailblazer API fixture"""

    return MockTB()



from typing import Optional

from cg.constants.subject import Gender, PlinkGender, RelationshipStatus


def set_parent_if_missing(parent: Optional[str]) -> str:
    return RelationshipStatus.HAS_NO_PARENT if parent is None else parent


def set_gender_if_other(gender: Optional[str]) -> str:
    return PlinkGender.UNKNOWN if gender == Gender.OTHER else gender



"""Interactions with the genotype tool"""

import logging

from cg.exc import CaseNotFoundError
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class GenotypeAPI:
    """Interface with Genotype app.

    The config should contain a 'genotype' key:

        { 'database': 'mysql://localhost:3306/database' }
    """

    def __init__(self, config: dict):
        self.process = Process(
            binary=config["genotype"]["binary_path"], config=config["genotype"]["config_path"]
        )
        self.dry_run = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set the dry run state"""
        self.dry_run = dry_run

    def upload(self, bcf_path: str, samples_sex: dict, force: bool = False) -> None:
        """Upload genotypes for a family of samples."""
        upload_parameters = ["load", str(bcf_path)]
        if force:
            upload_parameters.append("--force")

        LOG.info("loading VCF genotypes for sample(s): %s", ", ".join(samples_sex.keys()))
        self.process.run_command(parameters=upload_parameters, dry_run=self.dry_run)

        for sample_id in samples_sex:
            # This is the sample sex specified by the customer
            sample_sex = samples_sex[sample_id]["pedigree"]
            self.update_sample_sex(sample_id, sample_sex)
            # This is the predicted sex based on variant calls from the pipeline
            analysis_predicted_sex = samples_sex[sample_id]["analysis"]
            self.update_analysis_sex(sample_id, sex=analysis_predicted_sex)

    def update_sample_sex(self, sample_id: str, sex: str) -> None:
        """Update the sex for a sample in the genotype tool"""
        sample_sex_parameters = ["add-sex", sample_id, "-s", sex]
        LOG.debug("Set sex for sample %s to %s", sample_id, sex)
        self.process.run_command(parameters=sample_sex_parameters, dry_run=self.dry_run)

    def update_analysis_sex(self, sample_id: str, sex: str) -> None:
        """Update the predicted sex for a sample based on genotype analysis in the genotype tool"""
        analysis_sex_parameters = ["add-sex", sample_id, "-a", "sequence", sex]
        LOG.debug("Set predicted sex for sample %s to %s for the sequence analysis", sample_id, sex)
        self.process.run_command(parameters=analysis_sex_parameters, dry_run=self.dry_run)

    def export_sample(self, days: int = 0) -> str:
        """Export sample info."""
        export_sample_parameters = ["export-sample", "-d", str(days)]

        self.process.run_command(parameters=export_sample_parameters, dry_run=self.dry_run)
        output = self.process.stdout
        # If sample not in genotype db, stdout of genotype command will be empty.
        if not output:
            raise CaseNotFoundError("samples not found in genotype db")
        return output

    def export_sample_analysis(self, days: int = 0) -> str:
        """Export analysis."""
        export_sample_analysis_parameters = ["export-sample-analysis", "-d", str(days)]

        self.process.run_command(parameters=export_sample_analysis_parameters, dry_run=self.dry_run)
        output = self.process.stdout
        # If sample not in genotype db, stdout of genotype command will be empty.
        if not output:
            raise CaseNotFoundError("samples not found in genotype db")
        return output

    def __str__(self):
        return f"GenotypeAPI(dry_run: {self.dry_run})"



from typing import Optional
from cg.store import Store
from cg.store.filters.status_metrics_filters import (
    filter_above_q30_threshold,
    filter_total_read_count_for_sample,
    filter_by_flow_cell_sample_internal_id_and_lane,
    filter_by_flow_cell_name,
    filter_by_sample_internal_id,
)
from cg.store.models import SampleLaneSequencingMetrics
from sqlalchemy.orm import Query
from tests.meta.demultiplex.conftest import flow_cell_name_demultiplexed_with_bcl_convert


def test_filter_total_read_count_for_sample(
    store_with_sequencing_metrics: Store, sample_id: str, expected_total_reads: int
):
    # GIVEN a Store with sequencing metrics
    metrics: Query = store_with_sequencing_metrics._get_query(table=SampleLaneSequencingMetrics)

    # WHEN getting total read counts for a sample
    total_reads_query: Query = filter_total_read_count_for_sample(
        metrics=metrics, sample_internal_id=sample_id
    )

    # THEN assert that the returned object is a Query
    assert isinstance(total_reads_query, Query)

    # THEN a total reads count is returned
    actual_total_reads: Optional[int] = total_reads_query.scalar()
    assert actual_total_reads

    # THEN assert that the actual total read count is as expected
    assert actual_total_reads == expected_total_reads


def test_filter_metrics_for_flow_cell_sample_internal_id_and_lane(
    store_with_sequencing_metrics: Store, sample_id: str, flow_cell_name: str
):
    # GIVEN a Store with sequencing metrics
    metrics: Query = store_with_sequencing_metrics._get_query(table=SampleLaneSequencingMetrics)

    # WHEN getting metrics for a flow cell, sample internal id and lane
    metrics_query: Query = filter_by_flow_cell_sample_internal_id_and_lane(
        metrics=metrics,
        flow_cell_name=flow_cell_name,
        sample_internal_id=sample_id,
        lane=1,
    )

    # THEN assert that the returned object is a Query
    assert isinstance(metrics_query, Query)

    # THEN assert that the query returns a list of metrics
    assert metrics_query.all()

    # THEN assert that the query returns the expected number of metrics
    assert len(metrics_query.all()) == 1

    # THEN assert that the query returns the expected metrics
    assert metrics_query[0].flow_cell_name == flow_cell_name
    assert metrics_query[0].sample_internal_id == sample_id
    assert metrics_query[0].flow_cell_lane_number == 1


def test_filter_metrics_by_flow_cell_name(
    store_with_sequencing_metrics: Store, flow_cell_name: str
):
    # GIVEN a Store with sequencing metrics
    metrics: Query = store_with_sequencing_metrics._get_query(table=SampleLaneSequencingMetrics)

    # WHEN getting metrics for a flow cell name
    metrics_query: Query = filter_by_flow_cell_name(metrics=metrics, flow_cell_name=flow_cell_name)

    # THEN assert that the returned object is a Query
    assert isinstance(metrics_query, Query)

    # THEN assert that the query returns a list of metrics
    assert metrics_query.all()

    # THEN assert that the query returns the expected number of metrics
    assert len(metrics_query.all()) == 1

    # THEN assert that the query returns the expected metrics
    for metric in metrics_query.all():
        assert metric.flow_cell_name == flow_cell_name


def test_filter_metrics_by_sample_internal_id(store_with_sequencing_metrics: Store, sample_id: str):
    # GIVEN a Store with sequencing metrics
    metrics: Query = store_with_sequencing_metrics._get_query(table=SampleLaneSequencingMetrics)

    # WHEN getting metrics for a sample internal id
    metrics_query: Query = filter_by_sample_internal_id(
        sample_internal_id=sample_id, metrics=metrics
    )

    # THEN assert that the returned object is a Query
    assert isinstance(metrics_query, Query)

    # THEN assert that the query returns a list of metrics
    assert metrics_query.all()

    # THEN assert that the query returns the expected number of metrics
    assert len(metrics_query.all()) == 2

    # THEN assert that the query returns the expected metrics
    for metric in metrics_query.all():
        assert metric.sample_internal_id == sample_id


def test_filter_above_q30_threshold(store_with_sequencing_metrics: Store):
    # GIVEN a Store with sequencing metrics
    metrics: Query = store_with_sequencing_metrics._get_query(table=SampleLaneSequencingMetrics)
    metric: Optional[SampleLaneSequencingMetrics] = metrics.first()
    assert metric

    # GIVEN a Q30 threshold that at least one metric will pass
    q30_threshold = int(metric.sample_base_percentage_passing_q30 / 2)

    # WHEN filtering metrics above the Q30 threshold
    filtered_metrics: Query = filter_above_q30_threshold(
        metrics=metrics,
        q30_threshold=q30_threshold,
    )

    # THEN assert that the returned object is a Query
    assert isinstance(filtered_metrics, Query)

    # THEN assert that the query returns a list of filtered metrics
    assert filtered_metrics.all()

    # THEN assert that all returned metrics have a sample_base_percentage_passing_q30 greater than the threshold
    for metric in filtered_metrics.all():
        assert metric.sample_base_percentage_passing_q30 > q30_threshold


def test_filter_metrics_by_flow_cell_name(
    store_with_sequencing_metrics: Store, flow_cell_name: str
):
    # GIVEN a Store with sequencing metrics
    metrics: Query = store_with_sequencing_metrics._get_query(table=SampleLaneSequencingMetrics)

    # WHEN filtering metrics by flow cell name
    filtered_metrics: Query = filter_by_flow_cell_name(
        metrics=metrics, flow_cell_name=flow_cell_name
    )

    # THEN assert that the returned object is a Query
    assert isinstance(filtered_metrics, Query)

    # THEN assert that the query returns a list of filtered metrics
    assert filtered_metrics.all()

    # THEN assert that all returned metrics have the expected flow cell name
    for metric in filtered_metrics.all():
        assert metric.flow_cell_name == flow_cell_name



from typing import List
from pydantic import BaseModel, Field


class ReadMetric(BaseModel):
    """
    Represents a set of metrics generated by bcl2fastq for a read in the sequencing process.
    """

    yield_: int = Field(..., alias="Yield", ge=0)
    yield_q30: int = Field(..., alias="YieldQ30", ge=0)
    quality_score_sum: int = Field(..., alias="QualityScoreSum", ge=0)


class DemuxResult(BaseModel):
    """
    Represents the result of the bcl2fastq demultiplexing process for a given sample.
    """

    sample_id: str = Field(..., alias="SampleId", min_length=1)
    number_reads: int = Field(..., alias="NumberReads", ge=0)
    yield_: int = Field(..., alias="Yield", ge=0)
    read_metrics: List[ReadMetric] = Field(..., alias="ReadMetrics")


class ConversionResult(BaseModel):
    """
    Represents the result of the conversion process for a given lane as generated by
    the bcl2fastq demultiplexing software.
    """

    lane_number: int = Field(..., alias="LaneNumber", gt=0)
    demux_results: List[DemuxResult] = Field(..., alias="DemuxResults")


class Bcl2FastqSampleLaneTileMetrics(BaseModel):
    """
    Represents a set of sequencing metrics for a bcl2fastq run per sample, lane and tile on a flow cell.
    """

    flow_cell_name: str = Field(..., alias="Flowcell", min_length=1)
    conversion_results: List[ConversionResult] = Field(..., alias="ConversionResults")


class Bcl2FastqSampleLaneMetrics(BaseModel):
    """
    Aggregated metrics per sample and lane from a bcl2fastq run.
    """

    flow_cell_name: str
    flow_cell_lane_number: int
    sample_id: str
    sample_total_reads_in_lane: int
    sample_total_yield_in_lane: int
    sample_total_yield_q30_in_lane: int
    sample_total_quality_score_in_lane: int



from typing import List, Generator

import pytest

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.meta.transfer import TransferLims
from cg.store import Store
from pathlib import Path

from tests.mocks.limsmock import MockLimsAPI


@pytest.fixture(name="transfer_lims_api")
def transfer_lims_api(sample_store: Store) -> Generator[TransferLims, None, None]:
    """Setup LIMS transfer API."""
    yield TransferLims(sample_store, MockLimsAPI(config=""))


@pytest.fixture(scope="session")
def external_data_directory(
    tmpdir_factory, customer_id: str, cust_sample_id: str, ticket_id: str
) -> Path:
    """Returns a customer folder with fastq.gz files in sample-directories."""
    cust_folder: Path = tmpdir_factory.mktemp(customer_id, numbered=False)
    ticket_folder: Path = Path(cust_folder, ticket_id)
    ticket_folder.mkdir()
    samples: List[str] = [f"{cust_sample_id}1", f"{cust_sample_id}2"]
    for sample in samples:
        Path(ticket_folder, sample).mkdir(exist_ok=True, parents=True)
        for read in [1, 2]:
            Path(ticket_folder, sample, f"{sample}_fastq_{read}.fastq.gz").touch(exist_ok=True)
            Path(ticket_folder, sample, f"{sample}_fastq_{read}.fastq.gz.md5").touch(exist_ok=True)
    return Path(ticket_folder)


@pytest.fixture(name="sample_sheet_path")
def sample_sheet_path(tmpdir_factory) -> Generator[Path, None, None]:
    """Create and return path to sample sheet."""
    sample_sheet_path_dir: Path = Path(tmpdir_factory.mktemp("DEMUX"), "HVKJCDRXX", "NAADM1")
    sample_sheet_path_dir.mkdir(parents=True, exist_ok=True)
    sample_sheet_path: Path = Path(
        sample_sheet_path_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )
    sample_sheet_path.touch()
    yield sample_sheet_path



from pathlib import Path
from typing import Dict, List
import pytest
from cg.store.models import Sample
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.meta.demultiplex.housekeeper_storage_functions import add_sample_sheet_path_to_housekeeper
from cg.models.cg_config import CGConfig
from tests.meta.demultiplex.conftest import FlowCellInfo


def test_set_dry_run(
    demultiplex_context: CGConfig,
):
    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)

    # THEN dry run should be False
    assert post_demux_api.dry_run is False

    # WHEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # THEN dry run should be True
    assert post_demux_api.dry_run is True


@pytest.mark.parametrize(
    "demux_type",
    ["BCL2FASTQ_TREE", "BCLCONVERT_TREE", "BCLCONVERT_FLAT", "BCLCONVERT_ON_SEQUENCER"],
)
def test_post_processing_of_flow_cell(
    demux_type: str,
    demultiplex_context: CGConfig,
    flow_cell_info_map: Dict[str, FlowCellInfo],
    tmp_demultiplexed_runs_directory: Path,
    novaseq6000_bcl_convert_sample_sheet_path,
):
    """Test adding a demultiplexed flow cell to the databases with. Runs on each type of
    demultiplexing software and setting used."""

    # GIVEN a demultiplexed flow cell
    flow_cell_demultiplexing_directory: str = flow_cell_info_map.get(demux_type).directory
    flow_cell_name: str = flow_cell_info_map.get(demux_type).name
    sample_internal_ids: List[str] = flow_cell_info_map.get(demux_type).sample_internal_ids

    # GIVEN the sample_internal_ids are present in statusdb
    for sample_internal_id in sample_internal_ids:
        assert demultiplex_context.status_db.get_sample_by_internal_id(sample_internal_id)

    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a directory with a flow cell demultiplexed with BCL Convert
    demux_post_processing_api.demultiplexed_runs_dir = tmp_demultiplexed_runs_directory

    # GIVEN that the sample sheet is in housekeeper
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=Path(
            tmp_demultiplexed_runs_directory, flow_cell_demultiplexing_directory
        ),
        flow_cell_name=flow_cell_name,
        hk_api=demux_post_processing_api.hk_api,
    )

    # THEN the sample sheet is in housekeeper
    assert demux_post_processing_api.hk_api.get_files(
        bundle=flow_cell_name, tags=[SequencingFileTag.SAMPLE_SHEET]
    ).all()
    # WHEN post-processing the demultiplexed flow cell
    demux_post_processing_api.finish_flow_cell(flow_cell_demultiplexing_directory)

    # THEN a flow cell was created in statusdb
    assert demux_post_processing_api.status_db.get_flow_cell_by_name(flow_cell_name)

    # THEN sequencing metrics were created for the flow cell
    assert demux_post_processing_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name
    )
    # THEN the read count was calculated for all samples in the flow cell directory
    for sample_internal_id in sample_internal_ids:
        sample: Sample = demux_post_processing_api.status_db.get_sample_by_internal_id(
            sample_internal_id
        )
        assert isinstance(sample.reads, int)

    # THEN a bundle was added to Housekeeper for the flow cell
    assert demux_post_processing_api.hk_api.bundle(flow_cell_name)

    # THEN a bundle was added to Housekeeper for each sample
    for sample_internal_id in sample_internal_ids:
        assert demux_post_processing_api.hk_api.bundle(sample_internal_id)

    # THEN a sample sheet was added to Housekeeper
    assert demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.SAMPLE_SHEET],
        bundle=flow_cell_name,
    ).all()

    # THEN sample fastq files were added to Housekeeper tagged with FASTQ and the flow cell name
    for sample_internal_id in sample_internal_ids:
        assert demux_post_processing_api.hk_api.get_files(
            tags=[SequencingFileTag.FASTQ, flow_cell_name],
            bundle=sample_internal_id,
        ).all()

    # THEN a delivery file was created in the flow cell directory
    delivery_path = Path(
        demux_post_processing_api.demultiplexed_runs_dir,
        flow_cell_demultiplexing_directory,
        DemultiplexingDirsAndFiles.DELIVERY,
    )

    assert delivery_path.exists()


def test_get_all_demultiplexed_flow_cell_out_dirs(
    demultiplex_context: CGConfig,
    tmp_demultiplexed_runs_directory: Path,
    tmp_demultiplexed_runs_bcl2fastq_directory: Path,
):
    """Test returning all flow cell directories from the demultiplexing run directory."""
    # GIVEN a demultiplex flow cell finished output directory that exist

    # GIVEN a demultiplex context
    demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)
    demux_api.demultiplexed_runs_dir = tmp_demultiplexed_runs_directory

    # WHEN calling get_all_demultiplexed_flow_cell_dirs
    demultiplexed_flow_cell_dirs: List[Path] = demux_api.get_all_demultiplexed_flow_cell_dirs()

    # THEN the demultiplexed flow cells run directories should be returned
    assert tmp_demultiplexed_runs_bcl2fastq_directory in demultiplexed_flow_cell_dirs



import re


def is_valid_sample_internal_id(sample_internal_id: str) -> bool:
    """
    Check if a sample internal id has the correct structure:
    starts with three letters followed by at least three digits.
    """
    return bool(re.search(r"^[A-Za-z]{3}\d{3}", sample_internal_id))



import logging
from typing import Optional

from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.gens import GensAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.loqus import LoqusdbAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants.observations import LoqusdbInstance
from cg.constants.priority import SlurmQos
from cg.store import Store
from pydantic.v1 import BaseModel, EmailStr, Field
from typing_extensions import Literal

LOG = logging.getLogger(__name__)


class Sequencers(BaseModel):
    hiseqx: str
    hiseqga: str
    novaseq: str


class EncryptionDirs(BaseModel):
    current: str
    legacy: str


class FlowCellRunDirs(Sequencers):
    pass


class BackupConfig(BaseModel):
    encrypt_dir: EncryptionDirs


class CleanDirs(BaseModel):
    sample_sheets_dir_name: str
    flow_cell_run_dirs: FlowCellRunDirs


class CleanConfig(BaseModel):
    flow_cells: CleanDirs


class SlurmConfig(BaseModel):
    account: str
    hours: Optional[int]
    mail_user: EmailStr
    memory: Optional[int]
    number_tasks: Optional[int]
    conda_env: Optional[str]
    qos: SlurmQos = SlurmQos.LOW


class HousekeeperConfig(BaseModel):
    database: str
    root: str


class DemultiplexConfig(BaseModel):
    run_dir: str
    out_dir: str
    slurm: SlurmConfig


class TrailblazerConfig(BaseModel):
    service_account: str
    service_account_auth_file: str
    host: str


class StatinaConfig(BaseModel):
    host: Optional[str]
    user: str
    key: str
    api_url: str
    upload_path: str
    auth_path: str


class CommonAppConfig(BaseModel):
    binary_path: str
    config_path: Optional[str]


class FluffyUploadConfig(BaseModel):
    user: str
    password: str
    host: str
    remote_path: str
    port: int


class FluffyConfig(CommonAppConfig):
    root_dir: str
    sftp: FluffyUploadConfig


class LimsConfig(BaseModel):
    host: str
    username: str
    password: str


class CrunchyConfig(BaseModel):
    conda_binary: Optional[str] = None
    cram_reference: str
    slurm: SlurmConfig


class MutaccAutoConfig(CommonAppConfig):
    padding: int = 300


class BalsamicConfig(CommonAppConfig):
    balsamic_cache: str
    bed_path: str
    binary_path: str
    conda_env: str
    loqusdb_path: str
    pon_path: str
    root: str
    slurm: SlurmConfig
    swegen_path: str


class MutantConfig(BaseModel):
    binary_path: str
    conda_binary: Optional[str] = None
    conda_env: str
    root: str


class MipConfig(BaseModel):
    conda_binary: Optional[str] = None
    conda_env: str
    mip_config: str
    pipeline: str
    root: str
    script: str


class RnafusionConfig(CommonAppConfig):
    root: str
    references: str
    binary_path: str
    pipeline_path: str
    conda_env: str
    compute_env: str
    profile: str
    conda_binary: Optional[str] = None
    launch_directory: str
    revision: str
    slurm: SlurmConfig
    tower_binary_path: str
    tower_pipeline: str


class TaxprofilerConfig(CommonAppConfig):
    root: str
    binary_path: str
    conda_env: str
    profile: str
    pipeline_path: str
    revision: str
    conda_binary: Optional[str] = None
    hostremoval_reference: str
    databases: str
    slurm: SlurmConfig
    tower_binary_path: str
    tower_pipeline: str


class MicrosaltConfig(BaseModel):
    binary_path: str
    conda_binary: Optional[str] = None
    conda_env: str
    queries_path: str
    root: str


class GisaidConfig(CommonAppConfig):
    submitter: str
    log_dir: str
    logwatch_email: EmailStr
    upload_password: str
    upload_cid: str


class DataDeliveryConfig(BaseModel):
    destination_path: str
    covid_destination_path: str
    covid_source_path = str
    covid_report_path: str
    account: str
    base_path: str
    mail_user: str


class EmailBaseSettings(BaseModel):
    sender_email: EmailStr
    smtp_server: str


class FOHMConfig(BaseModel):
    host: str
    port: int
    key: str
    username: str
    valid_uploader: str
    email_sender: str
    email_recipient: str
    email_host: str


class ExternalConfig(BaseModel):
    hasta: str
    caesar: str


class DataFlowConfig(BaseModel):
    database_name: str
    user: str
    password: str
    url: str
    local_storage: str
    archive_repository: str


class CGConfig(BaseModel):
    database: str
    environment: Literal["production", "stage"] = "stage"
    madeline_exe: str
    delivery_path: str
    max_flowcells: Optional[int]
    email_base_settings: EmailBaseSettings
    flow_cells_dir: str
    demultiplexed_flow_cells_dir: str

    # Base APIs that always should exist
    status_db_: Store = None
    housekeeper: HousekeeperConfig
    housekeeper_api_: HousekeeperAPI = None

    # App APIs that can be instantiated in CGConfig
    backup: BackupConfig = None
    chanjo: CommonAppConfig = None
    chanjo_api_: ChanjoAPI = None
    clean: Optional[CleanConfig] = None
    crunchy: CrunchyConfig = None
    crunchy_api_: CrunchyAPI = None
    data_delivery: DataDeliveryConfig = Field(None, alias="data-delivery")
    data_flow_config: Optional[DataFlowConfig] = None
    demultiplex: DemultiplexConfig = None
    demultiplex_api_: DemultiplexingAPI = None
    encryption: Optional[CommonAppConfig] = None
    external: ExternalConfig = None
    genotype: CommonAppConfig = None
    genotype_api_: GenotypeAPI = None
    gens: CommonAppConfig = None
    gens_api_: GensAPI = None
    hermes: CommonAppConfig = None
    hermes_api_: HermesApi = None
    lims: LimsConfig = None
    lims_api_: LimsAPI = None
    loqusdb: CommonAppConfig = Field(None, alias=LoqusdbInstance.WGS.value)
    loqusdb_api_: LoqusdbAPI = None
    loqusdb_wes: CommonAppConfig = Field(None, alias=LoqusdbInstance.WES.value)
    loqusdb_somatic: CommonAppConfig = Field(None, alias=LoqusdbInstance.SOMATIC.value)
    loqusdb_tumor: CommonAppConfig = Field(None, alias=LoqusdbInstance.TUMOR.value)
    madeline_api_: MadelineAPI = None
    mutacc_auto: MutaccAutoConfig = Field(None, alias="mutacc-auto")
    mutacc_auto_api_: MutaccAutoAPI = None
    pdc: Optional[CommonAppConfig] = None
    scout: CommonAppConfig = None
    scout_api_: ScoutAPI = None
    tar: Optional[CommonAppConfig] = None
    trailblazer: TrailblazerConfig = None
    trailblazer_api_: TrailblazerAPI = None

    # Meta APIs that will use the apps from CGConfig
    balsamic: BalsamicConfig = None
    statina: StatinaConfig = None
    fohm: Optional[FOHMConfig] = None
    fluffy: FluffyConfig = None
    microsalt: MicrosaltConfig = None
    gisaid: GisaidConfig = None
    mip_rd_dna: MipConfig = Field(None, alias="mip-rd-dna")
    mip_rd_rna: MipConfig = Field(None, alias="mip-rd-rna")
    mutant: MutantConfig = None
    rnafusion: RnafusionConfig = Field(None, alias="rnafusion")
    taxprofiler: TaxprofilerConfig = Field(None, alias="taxprofiler")

    # These are meta APIs that gets instantiated in the code
    meta_apis: dict = {}

    class Config:
        arbitrary_types_allowed = True
        fields = {
            "chanjo_api_": "chanjo_api",
            "crunchy_api_": "crunchy_api",
            "demultiplex_api_": "demultiplex_api",
            "genotype_api_": "genotype_api",
            "gens_api_": "gens_api",
            "hermes_api_": "hermes_api",
            "housekeeper_api_": "housekeeper_api",
            "lims_api_": "lims_api",
            "loqusdb_api_": "loqusdb_api",
            "madeline_api_": "madeline_api",
            "mutacc_auto_api_": "mutacc_auto_api",
            "scout_api_": "scout_api",
            "status_db_": "status_db",
            "trailblazer_api_": "trailblazer_api",
        }

    @property
    def chanjo_api(self) -> ChanjoAPI:
        api = self.__dict__.get("chanjo_api_")
        if api is None:
            LOG.debug("Instantiating chanjo api")
            api = ChanjoAPI(config=self.dict())
            self.chanjo_api_ = api
        return api

    @property
    def crunchy_api(self) -> CrunchyAPI:
        api = self.__dict__.get("crunchy_api_")
        if api is None:
            LOG.debug("Instantiating crunchy api")
            api = CrunchyAPI(config=self.dict())
            self.crunchy_api_ = api
        return api

    @property
    def demultiplex_api(self) -> DemultiplexingAPI:
        demultiplex_api = self.__dict__.get("demultiplex_api_")
        if demultiplex_api is None:
            LOG.debug("Instantiating demultiplexing api")
            demultiplex_api = DemultiplexingAPI(
                config=self.dict(), housekeeper_api=self.housekeeper_api
            )
            self.demultiplex_api_ = demultiplex_api
        return demultiplex_api

    @property
    def genotype_api(self) -> GenotypeAPI:
        api = self.__dict__.get("genotype_api_")
        if api is None:
            LOG.debug("Instantiating genotype api")
            api = GenotypeAPI(config=self.dict())
            self.genotype_api_ = api
        return api

    @property
    def gens_api(self) -> GensAPI:
        """Returns Gens API after making sure it has been instantiated."""
        api = self.__dict__.get("gens_api_")
        if api is None:
            LOG.debug("Instantiating gens api")
            api = GensAPI(config=self.dict())
            self.gens_api_ = api
        return api

    @property
    def hermes_api(self) -> HermesApi:
        hermes_api = self.__dict__.get("hermes_api_")
        if hermes_api is None:
            LOG.debug("Instantiating hermes api")
            hermes_api = HermesApi(config=self.dict())
            self.hermes_api_ = hermes_api
        return hermes_api

    @property
    def housekeeper_api(self) -> HousekeeperAPI:
        housekeeper_api = self.__dict__.get("housekeeper_api_")
        if housekeeper_api is None:
            LOG.debug("Instantiating housekeeper api")
            housekeeper_api = HousekeeperAPI(config=self.dict())
            self.housekeeper_api_ = housekeeper_api
        return housekeeper_api

    @property
    def lims_api(self) -> LimsAPI:
        api = self.__dict__.get("lims_api_")
        if api is None:
            LOG.debug("Instantiating lims api")
            api = LimsAPI(config=self.dict())
            self.lims_api_ = api
        return api

    @property
    def loqusdb_api(self) -> LoqusdbAPI:
        api = self.__dict__.get("loqusdb_api_")
        if api is None:
            LOG.debug("Instantiating loqusdb api")
            api: LoqusdbAPI = LoqusdbAPI(
                binary_path=self.loqusdb.binary_path, config_path=self.loqusdb.config_path
            )
            self.loqusdb_api_ = api
        return api

    @property
    def madeline_api(self) -> MadelineAPI:
        api = self.__dict__.get("madeline_api_")
        if api is None:
            LOG.debug("Instantiating madeline api")
            api = MadelineAPI(config=self.dict())
            self.madeline_api_ = api
        return api

    @property
    def mutacc_auto_api(self) -> MutaccAutoAPI:
        api = self.__dict__.get("mutacc_auto_api_")
        if api is None:
            LOG.debug("Instantiating mutacc_auto api")
            api = MutaccAutoAPI(config=self.dict())
            self.mutacc_auto_api_ = api
        return api

    @property
    def scout_api(self) -> ScoutAPI:
        api = self.__dict__.get("scout_api_")
        if api is None:
            LOG.debug("Instantiating scout api")
            api = ScoutAPI(config=self.dict())
            self.scout_api_ = api
        return api

    @property
    def status_db(self) -> Store:
        status_db = self.__dict__.get("status_db_")
        if status_db is None:
            LOG.debug("Instantiating status db")
            status_db = Store(uri=self.database)
            self.status_db_ = status_db
        return status_db

    @property
    def trailblazer_api(self) -> TrailblazerAPI:
        api = self.__dict__.get("trailblazer_api_")
        if api is None:
            LOG.debug("Instantiating trailblazer api")
            api = TrailblazerAPI(config=self.dict())
            self.trailblazer_api_ = api
        return api



from pathlib import Path

import pytest
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.osticket import MockOsTicket

from cg.meta.orders import OrdersAPI
from cg.meta.orders.api import FastqSubmitter
from cg.meta.orders.balsamic_submitter import BalsamicSubmitter
from cg.meta.orders.metagenome_submitter import MetagenomeSubmitter
from cg.meta.orders.microbial_submitter import MicrobialSubmitter
from cg.meta.orders.mip_dna_submitter import MipDnaSubmitter
from cg.meta.orders.mip_rna_submitter import MipRnaSubmitter
from cg.meta.orders.rml_submitter import RmlSubmitter
from cg.meta.orders.ticket_handler import TicketHandler
from cg.models.orders.order import OrderIn, OrderType
from cg.store import Store
from tests.apps.orderform.conftest import (
    balsamic_order_to_submit,
    fastq_order_to_submit,
    metagenome_order_to_submit,
    microbial_order_to_submit,
    mip_order_to_submit,
    mip_rna_order_to_submit,
    rml_order_to_submit,
    rnafusion_order_to_submit,
    sarscov2_order_to_submit,
)


@pytest.fixture(scope="session")
def all_orders_to_submit(
    balsamic_order_to_submit: dict,
    fastq_order_to_submit: dict,
    metagenome_order_to_submit: dict,
    microbial_order_to_submit: dict,
    mip_order_to_submit: dict,
    mip_rna_order_to_submit: dict,
    rml_order_to_submit: dict,
    rnafusion_order_to_submit: dict,
    sarscov2_order_to_submit: dict,
) -> dict:
    """Returns a dict of parsed order for each order type."""
    return {
        OrderType.BALSAMIC: OrderIn.parse_obj(balsamic_order_to_submit, project=OrderType.BALSAMIC),
        OrderType.FASTQ: OrderIn.parse_obj(fastq_order_to_submit, project=OrderType.FASTQ),
        OrderType.FLUFFY: OrderIn.parse_obj(rml_order_to_submit, project=OrderType.FLUFFY),
        OrderType.METAGENOME: OrderIn.parse_obj(
            metagenome_order_to_submit, project=OrderType.METAGENOME
        ),
        OrderType.MICROSALT: OrderIn.parse_obj(
            microbial_order_to_submit, project=OrderType.MICROSALT
        ),
        OrderType.MIP_DNA: OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA),
        OrderType.MIP_RNA: OrderIn.parse_obj(mip_rna_order_to_submit, project=OrderType.MIP_RNA),
        OrderType.RML: OrderIn.parse_obj(rml_order_to_submit, project=OrderType.RML),
        OrderType.RNAFUSION: OrderIn.parse_obj(
            rnafusion_order_to_submit, project=OrderType.RNAFUSION
        ),
        OrderType.SARS_COV_2: OrderIn.parse_obj(
            sarscov2_order_to_submit, project=OrderType.SARS_COV_2
        ),
    }


@pytest.fixture
def balsamic_status_data(balsamic_order_to_submit: dict):
    """Parse balsamic order example."""
    project: OrderType = OrderType.BALSAMIC
    order: OrderIn = OrderIn.parse_obj(obj=balsamic_order_to_submit, project=project)
    return BalsamicSubmitter.order_to_status(order=order)


@pytest.fixture
def fastq_status_data(fastq_order_to_submit):
    """Parse fastq order example."""
    project: OrderType = OrderType.FASTQ
    order: OrderIn = OrderIn.parse_obj(obj=fastq_order_to_submit, project=project)
    return FastqSubmitter.order_to_status(order=order)


@pytest.fixture
def metagenome_status_data(metagenome_order_to_submit: dict):
    """Parse metagenome order example."""
    project: OrderType = OrderType.METAGENOME
    order: OrderIn = OrderIn.parse_obj(obj=metagenome_order_to_submit, project=project)

    return MetagenomeSubmitter.order_to_status(order=order)


@pytest.fixture
def microbial_status_data(microbial_order_to_submit: dict):
    """Parse microbial order example."""
    project: OrderType = OrderType.MICROSALT
    order: OrderIn = OrderIn.parse_obj(obj=microbial_order_to_submit, project=project)
    return MicrobialSubmitter.order_to_status(order=order)


@pytest.fixture
def mip_rna_status_data(mip_rna_order_to_submit: dict):
    """Parse rna order example."""
    project: OrderType = OrderType.MIP_RNA
    order: OrderIn = OrderIn.parse_obj(obj=mip_rna_order_to_submit, project=project)
    return MipRnaSubmitter.order_to_status(order=order)


@pytest.fixture
def mip_status_data(mip_order_to_submit: dict):
    """Parse scout order example."""
    project: OrderType = OrderType.MIP_DNA
    order: OrderIn = OrderIn.parse_obj(obj=mip_order_to_submit, project=project)
    return MipDnaSubmitter.order_to_status(order=order)


@pytest.fixture
def rml_status_data(rml_order_to_submit):
    """Parse rml order example."""
    project: OrderType = OrderType.RML
    order: OrderIn = OrderIn.parse_obj(obj=rml_order_to_submit, project=project)
    return RmlSubmitter.order_to_status(order=order)


@pytest.fixture(scope="function")
def orders_api(base_store, osticket: MockOsTicket, lims_api: MockLimsAPI):
    return OrdersAPI(lims=lims_api, status=base_store, osticket=osticket)


@pytest.fixture
def ticket_handler(store: Store, osticket: MockOsTicket) -> TicketHandler:
    return TicketHandler(status_db=store, osticket_api=osticket)



"""Module for Rnafusion analysis API tests."""
from typing import List

from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase
from cg.models.rnafusion.rnafusion import RnafusionAnalysis


def test_parse_analysis(
    rnafusion_context: CGConfig,
    rnafusion_case_id: str,
    sample_id: str,
    rnafusion_multiqc_json_metrics: dict,
    rnafusion_metrics: dict,
    mock_analysis_finish,
):
    """Test Rnafusion output analysis files parsing."""

    # GIVEN a Rnafusion analysis API and a list of QC metrics
    analysis_api: RnafusionAnalysisAPI = rnafusion_context.meta_apis["analysis_api"]
    qc_metrics: List[MetricsBase] = analysis_api.get_multiqc_json_metrics(case_id=rnafusion_case_id)

    # WHEN extracting the analysis model
    analysis_model: RnafusionAnalysis = analysis_api.parse_analysis(qc_metrics_raw=qc_metrics)

    # THEN the analysis model and its content should have been correctly extracted
    assert analysis_model.sample_metrics[sample_id] == rnafusion_metrics


def test_get_latest_metadata(
    rnafusion_context: CGConfig, rnafusion_case_id: str, mock_analysis_finish
):
    """Test retrieval of Rnafusion latest metadata."""

    # GIVEN a Rnafusion analysis API and a list of QC metrics
    analysis_api: RnafusionAnalysisAPI = rnafusion_context.meta_apis["analysis_api"]

    # WHEN collecting the latest metadata
    latest_metadata: RnafusionAnalysis = analysis_api.get_latest_metadata(case_id=rnafusion_case_id)

    # THEN the latest metadata should have been parsed
    assert latest_metadata
    assert latest_metadata.sample_metrics



from typing import List, Tuple

import pytest
from cg.apps.demultiplex.sample_sheet.index import Index
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreatorBcl2Fastq,
    SampleSheetCreatorBCLConvert,
)
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


@pytest.fixture
def bcl_convert_samples_with_updated_indexes() -> List[FlowCellSampleBCLConvert]:
    """Return a list of three FlowCellSampleBCLConvert with updated indexes."""
    sample_1 = FlowCellSampleBCLConvert(
        lane=1, sample_id="sample 1", index="CAGAAGAT", index2="CAATGTAC"
    )
    sample_2 = FlowCellSampleBCLConvert(
        lane=1, sample_id="sample 2", index="CAGAAGAG", index2="CAATGTAT"
    )
    sample_3 = FlowCellSampleBCLConvert(
        lane=2, sample_id="sample 3", index="AAGCGATAGA", index2="AACCGCAACA"
    )
    return [sample_1, sample_2, sample_3]


@pytest.fixture
def override_cycles_for_samples_with_updated_indexes() -> List[str]:
    """Return the correspondent Override Cycles values for three samples."""
    return ["Y151;I8N2;N2I8;Y151", "Y151;I8N2;N2I8;Y151", "Y151;I10;I10;Y151"]


@pytest.fixture
def barcode_mismatch_values_for_samples_with_updated_indexes() -> List[Tuple[int, int]]:
    """Return the pairs of barcode mismatch values corresponding to three samples."""
    return [(0, 0), (0, 0), (1, 1)]


@pytest.fixture
def valid_index() -> Index:
    """Return a valid index."""
    return Index(name="C07 - UDI0051", sequence="AACAGGTT-ATACCAAG")


@pytest.fixture
def bcl2fastq_sample_sheet_creator(
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    lims_novaseq_bcl2fastq_samples: List[FlowCellSampleBcl2Fastq],
) -> SampleSheetCreatorBcl2Fastq:
    """Returns a sample sheet creator for version 1 sample sheets with bcl2fastq format."""
    return SampleSheetCreatorBcl2Fastq(
        flow_cell=bcl2fastq_flow_cell,
        lims_samples=lims_novaseq_bcl2fastq_samples,
    )


@pytest.fixture
def bcl_convert_sample_sheet_creator(
    bcl_convert_flow_cell: FlowCellDirectoryData,
    lims_novaseq_bcl_convert_samples: List[FlowCellSampleBCLConvert],
) -> SampleSheetCreatorBCLConvert:
    """Returns a sample sheet creator for version 2 sample sheets with dragen format."""
    return SampleSheetCreatorBCLConvert(
        flow_cell=bcl_convert_flow_cell,
        lims_samples=lims_novaseq_bcl_convert_samples,
    )


# Sample sheet validation


@pytest.fixture
def sample_sheet_line_sample_1() -> List[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A68",
        "hg19",
        "ATTCCACACT",
        "TGGTCTTGTT",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture
def sample_sheet_line_sample_2() -> List[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A1",
        "hg19",
        "AGTTAGCTGG",
        "GATGAGAATG",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_data_header() -> List[List[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBcl2FastqSections.Data.LANE.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value,
            SampleSheetBcl2FastqSections.Data.INDEX_1.value,
            SampleSheetBcl2FastqSections.Data.INDEX_2.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value,
            SampleSheetBcl2FastqSections.Data.CONTROL.value,
            SampleSheetBcl2FastqSections.Data.RECIPE.value,
            SampleSheetBcl2FastqSections.Data.OPERATOR.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


@pytest.fixture
def sample_sheet_samples_no_header(
    sample_sheet_line_sample_1: List[str], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a sample sheet with samples but without a sample header."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def valid_sample_sheet_bcl2fastq(
    sample_sheet_bcl2fastq_data_header: List[List[str]],
    sample_sheet_line_sample_1: List[str],
    sample_sheet_line_sample_2: List[str],
) -> List[List[str]]:
    """Return the content of a valid Bcl2fastq sample sheet."""
    return sample_sheet_bcl2fastq_data_header + [
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_duplicate_same_lane(
    valid_sample_sheet_bcl2fastq: List[List[str]], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_bcl2fastq.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_bcl2fastq


@pytest.fixture
def sample_sheet_bcl2fastq_duplicate_different_lane(
    valid_sample_sheet_bcl2fastq: List[List[str]],
) -> List[List[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_bcl2fastq.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_bcl2fastq


@pytest.fixture
def valid_sample_sheet_dragen(
    sample_sheet_line_sample_1: List[str], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a valid Dragen sample sheet."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            "Lane",
            "Sample_ID",
            "SampleRef",
            "index",
            "index2",
            "SampleName",
            "Control",
            "Recipe",
            "Operator",
            "Sample_Project",
        ],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def sample_sheet_dragen_duplicate_same_lane(
    valid_sample_sheet_dragen: List[List[str]], sample_sheet_line_sample_2: List[str]
) -> List[List[str]]:
    """Return the content of a Dragen sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_dragen.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_dragen


@pytest.fixture
def sample_sheet_dragen_duplicate_different_lane(
    valid_sample_sheet_dragen: List[List[str]],
) -> List[List[str]]:
    """Return the content of a Dragen sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_dragen.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_dragen


@pytest.fixture
def novaseq6000_flow_cell_sample_1() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=1,
        SampleID="ACC7628A68",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture(name="novaseq6000_flow_cell_sample_2")
def novaseq6000_flow_cell_sample_2() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture(name="novaseq_x_flow_cell_sample_before_adapt_indexes")
def novaseq_x_flow_cell_sample_before_adapt_indexes() -> FlowCellSampleBCLConvert:
    """Return a NovaSeqX sample."""
    return FlowCellSampleBCLConvert(
        Lane=2,
        Sample_ID="ACC7628A1",
        index="ATTCCACACT-TGGTCTTGTT",
    )


@pytest.fixture(name="novaseq6000_flow_cell_sample_no_dual_index")
def novaseq6000_flow_cell_sample_no_dual_index() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT",
        SampleName="814206",
        Project="814206",
    )


@pytest.fixture(name="novaseq6000_flow_cell_sample_before_adapt_indexes")
def novaseq6000_flow_cell_sample_before_adapt_indexes() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT-TGGTCTTGTT",
        SampleName="814206",
        Project="814206",
    )



import logging
import os
from pathlib import Path
from typing import List

from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import (
    Bcl2FastqSampleLaneMetrics,
    Bcl2FastqSampleLaneTileMetrics,
)
from cg.constants.demultiplexing import (
    BCL2FASTQ_METRICS_DIRECTORY_NAME,
    BCL2FASTQ_METRICS_FILE_NAME,
)

LOG = logging.getLogger(__name__)


def parse_bcl2fastq_sequencing_metrics(flow_cell_dir: Path) -> List[Bcl2FastqSampleLaneMetrics]:
    """Parse metrics for a flow cell demultiplexed with Bcl2fastq."""
    tile_sequencing_metrics: List[
        Bcl2FastqSampleLaneTileMetrics
    ] = parse_bcl2fastq_raw_tile_metrics(flow_cell_dir)

    return aggregate_tile_metrics_per_sample_and_lane(tile_sequencing_metrics)


def parse_bcl2fastq_raw_tile_metrics(
    demultiplex_result_directory: Path,
) -> List[Bcl2FastqSampleLaneTileMetrics]:
    """Parse metrics for each tile on a flow cell demultiplexed with Bcl2fastq."""
    tile_sequencing_metrics: List[Bcl2FastqSampleLaneTileMetrics] = []

    stats_json_paths: List[Path] = get_bcl2fastq_stats_paths(
        demultiplex_result_directory=demultiplex_result_directory
    )

    for stats_json_path in stats_json_paths:
        LOG.debug(f"Parsing stats.json file {stats_json_path}")
        sequencing_metrics = Bcl2FastqSampleLaneTileMetrics.parse_file(stats_json_path)
        tile_sequencing_metrics.append(sequencing_metrics)

    return tile_sequencing_metrics


def aggregate_tile_metrics_per_sample_and_lane(
    tile_metrics: List[Bcl2FastqSampleLaneTileMetrics],
) -> List[Bcl2FastqSampleLaneMetrics]:
    """Aggregate the tile metrics per sample and lane instead."""
    metrics = {}

    for tile_metric in tile_metrics:
        for conversion_result in tile_metric.conversion_results:
            for demux_result in conversion_result.demux_results:
                sample_lane_key = (
                    conversion_result.lane_number,
                    demux_result.sample_id,
                )

                sample_id: str = discard_index_sequence(sample_id_with_index=demux_result.sample_id)

                if sample_lane_key not in metrics:
                    metrics[sample_lane_key] = Bcl2FastqSampleLaneMetrics(
                        flow_cell_name=tile_metric.flow_cell_name,
                        flow_cell_lane_number=conversion_result.lane_number,
                        sample_id=sample_id,
                        sample_total_reads_in_lane=0,
                        sample_total_yield_in_lane=0,
                        sample_total_yield_q30_in_lane=0,
                        sample_total_quality_score_in_lane=0,
                    )
                # Double the total reads since they are reported in pairs
                metrics[sample_lane_key].sample_total_reads_in_lane += demux_result.number_reads * 2
                metrics[sample_lane_key].sample_total_yield_in_lane += demux_result.yield_
                metrics[sample_lane_key].sample_total_yield_q30_in_lane += sum(
                    [read_metric.yield_q30 for read_metric in demux_result.read_metrics]
                )
                metrics[sample_lane_key].sample_total_quality_score_in_lane += sum(
                    [read_metric.quality_score_sum for read_metric in demux_result.read_metrics]
                )

    return list(metrics.values())


def discard_index_sequence(sample_id_with_index: str) -> str:
    """Discard the index sequence from the sample id."""
    return sample_id_with_index.split("_")[0]


def get_bcl2fastq_stats_paths(demultiplex_result_directory: Path) -> List[Path]:
    """
    Finds metrics files in Bcl2fastq demultiplex result directory.
    Raises:
        FileNotFoundError: If no stats.json files are found in the demultiplex result directory.
    """
    stats_json_paths = []

    for root, _, files in os.walk(demultiplex_result_directory):
        if root.endswith(BCL2FASTQ_METRICS_DIRECTORY_NAME) and BCL2FASTQ_METRICS_FILE_NAME in files:
            stats_json_paths.append(Path(root, BCL2FASTQ_METRICS_FILE_NAME))

    if not stats_json_paths:
        raise FileNotFoundError(
            f"Could not find any stats.json files in {demultiplex_result_directory}"
        )

    return stats_json_paths
